/*
        FileName: Copy4.cc
         Version: 2.0
     Written for: Visual dBASE 7.x
          Author: Gary White (see acknowledgement below)
            Date: May 17, 1999

 Acknowledgement: This code is based on the work of Christer Persson,
                  who did the initial develpment of a program called
                  db4copy.prg.  If you wonder why the file was renamed, 
                  I contacted Christer regarding my proposed changes and 
                  he indicated that he would be very happy if someone 
                  could develop the program further, but he was involved 
                  in some other things at the moment, didn't really have 
                  the time to work on it, and didn't think he would get 
                  back to it unless he had a particular need for it. He 
                  asked that revisions be renamed to avoid confusion with 
                  the original.

                  This version uses Christer's code to create an empty
                  table with a single dummy field.  It then uses the
                  LocalSQL ALTER TABLE command to add the required fields
                  and delete the dummy field.  This version also uses two
                  query objects to copy the data from the source table to 
                  the new destination table.  This method overcomes the
                  shortcoming of not being able to copy memo fields and
                  also cuts execution time roughly in half.

			License:	This code is FREEWARE. You are free to include this 
         			code in your own dBASE programs and distribute it 
                  freely as part of your programs as object code.

                  You may freely distribute this code to other developers 
                  for inclusion in their programs.

                  You may NOT sell this code or use it in any way whereby 
                  it is NOT included in a program.

                  You may make any revisions desired to this code.

                  *******************************************************
                  *                                                     *
                  *  Warning: This code is not warranteed in any way    *
                  *  whatsoever.  Use it at your own risk. Your use of  *
                  *	this code releases Gary White from any liability   *
                  *	whatsoever as regards the use, suitability or      *
                  *	quality of this code.                              *
                  *                                                     *
                  *******************************************************


 Version history: 1998-07-25 - db4copy.prg 
 						Original version 1.00 (Christer Persson)

                  1999-03-07 - db4copy.prg
                  Version 1.01 (Christer Persson)
                  Added this.recordNo=0 in method copyDB4.
                  This fixes a problem with wrong number
                  of records in destination file when more than
                  one file is copied. Thank you to Steve Tipton.
                  
                  1999-05-14 - Copy4.cc
                  Version 2.0 (Gary White)
                  Renamed from db4copy.prg to Copy4.cc.
                  Fixed bug with _DBASELOCK field being copied.
                  Added language driver defines header file.
                  Changed to using low-level file writing to create
   	               the empty table with a single field and LocalSQL
	                  to add and delete fields.
                  Changed the WriteRow method to use 2 query
	                  objects instead of low level file write.
                  Added support for TimeStamp fields.
                  Added support for Memo fields.
                  Minor code cleanup.


           Notes: Field type mappings are as follows

                  Level 7          Level 4
                  -------          -------
                  CHARACTER        CHARACTER
                  NUMERIC          NUMERIC
                  LOGICAL          LOGICAL
                  DATE             DATE
                  FLOAT            NUMERIC
                  LONG             NUMERIC
                  TIMESTAMP        CHARACTER
                  DOUBLE           NUMERIC
                  AUTOINCREMENT    NUMERIC
                  MEMO             MEMO
                  OLE              Not copied
                  BINARY           Not copied

                  Note that all numeric data types (numeric, float,
                  long, double, and autoincrement) are converted to
                  numeric. While dBASE IV did include a FLOAT field
                  type, using FLOAT with LocalSQL, creates a DOUBLE
                  field type which forces conversion back to a 
                  level 7 table.

                  TimeStamp fields are converted to CHARACTER fields
                  using the date object's ToLocaleString method, 
                  which will respect local date format settings.  It
                  also respects the setting of SET CENTURY, so if 
                  you need 4 digit years, you will need to ensure 
                  SET CENTURY ON.
*/



// language drivers
#include db4copy.h


#define showError(m)	msgbox(m,'Error',16)
#define confirm {|q,t|msgbox(q,t,32+4+256)==6}
#define atrim {|x|ltrim(trim(x))}



//======================================================================

function Copy4

/*======================================================================

        Function: Copy4

           Usage: Copy4( <oRowset>, <cFilename> [, <bShowProg>] )

         Example: Copy4( form.rowset, "c:\MyData\NewName.dbf" )

      Parameters:	oRowset     Object reference to the rowset to copy

                  cFilename   Name of the level 4 table to create

                  bShowProg   Optional boolean value:
                              TRUE (default), progress form is shown
                              FALSE, progress form is not displayed

           Calls: None

         Returns: True if successful, False if it fails to complete.

     Description: Creates a level 4 dbf that may be used to export data
                  to other applications.

*/
parameters rSrc, cDest, bShowProg

local bRetVal, qSrc, d

bRetVal = false

if argcount() < 2 or ;
	type('rSrc.endOfSet') # 'L' or ;
	type('cDest') # 'C'
   ShowError('Missing, or invalid parameters')
   return bRetVal
endif


try
	if right(upper(cDest),4) # '.DBF'
   	cDest += '.dbf'
   endif
   if file( cDest )
   	if not confirm('Overwrite existing file: '+cDest+'?','File exists')
      	return bRetVal
      endif
   endif

   // work with a copy of the rowset so we don't disturb anything
   // including record pointers.
   qSrc=new query()
   qSrc.database = rSrc.parent.database
   qSrc.sql=rSrc.parent.sql
   qSrc.active=true

   d=new dbf4()
   d.source = qSrc.rowset
   d.destination=cDest
	d.ShowProgress = iif( ArgCount() < 3, true, bShowProg )

   // try the copy
   bRetVal := d.MakeCopy()

   // clean up
   release object d
   qSrc.active=false
   release object qSrc
   release qSrc

catch (exception e)
	ShowError( e.message )
   bRetVal := false
endtry
return bRetVal




//======================================================================

class dbf4

/*======================================================================
	Custom class for exporting data to level 4 DBF file.

           Usage: set procedure to copy4.cc additive
           			dbf = new dbf4()
                  dbf.source := form.rowset
                  dbf.destination := "c:\MyData\NewName.dbf"
                  b = dbf.MakeCopy()
                  ? "Operation was "+iif(not b,"not",""),"completed"
                  ? dbf.RecordsCopied,"records were copied"

      Properties:

      				source	Rowset object to copy

             destination	String containing the name and, optionally,
             					the path of the new file to create.

            showProgress	Boolean value that determines whether or not
            					a progress form is displayed during the copy.
                           Default value is TRUE.

            	  LDriver	Language driver to use in the new file.
                 				See the include file, Copy4.h for details.

           RecordsCopied	The number of records successfully copied


      Methods:

      			 MakeCopy	Makes the level 4 table an copies records

                 Release	Releases the object

*/
	protect headerSize, recordSize, recordNo, ;
   	fileObj, aFields, aNames, ProgForm, ;
      addFields, copyData, createFile, db4Name, ;
      write16, write32, CopyToLevel4, ReleaseMe

   // published properties
	this.source = null				// source rowset to copy from
   this.destination = ''			// [path] filename to copy to
   this.ShowProgress = true		// whether or not to display progress form
   this.ldriver = WEurope_ANSI	// language driver (see copy4.h)
   this.RecordsCopied = 0			// number of records copied

   // protected properties
   this.headerSize = 0
   this.recordSize = 0
   this.recordNo = 0
   this.aFields = null
   this.aNames = null
   this.progForm = null
   this.fileObj = new file()
   this.MakeCopy = class::CopyTolevel4
   this.release = class::ReleaseMe

   function CopyToLevel4
   	local bRetVal
      private sEsc
      bRetVal = false	// in case of cancel before getting started
      try
      	_app.cancelCopy = false		// initialize flag
         sEsc = set('ON ESCAPE')		// save old ON ESCAPE setting
         on escape _app.cancelCopy := true	// allow user to cancel

         // Here, we'll create the progress form regardless of whether
         // it is displayed, or not.  That way we can assign values
         // to the progress object without having to test to see if
         // it is displayed while inside the loop.
         this.progForm = new progForm()	// create the progress form
         if this.ShowProgress
	         this.progForm.open()
         endif
		   bRetVal = true		// default to success
         if this.createFile() and not _app.cancelCopy
         	if this.addFields() and not _app.cancelCopy
         		bRetVal := this.copyData() and not _app.cancelCopy
            endif
         endif
      catch (exception e)
         ShowError( e.message )
         bRetVal := false
      finally
      	if type('this.progForm') == 'O'
         	this.progForm.release()
         endif
      endtry
      bRetVal = bRetVal and not _app.cancelCopy
      _app.cancelCopy := null
      on escape &sEsc
      return bRetVal

	function Releaseme
   	release object this.fileObj
      release object this
      return

   function createFile
   	// creates a level 4 dbf with one dummy field
      // named XDELETEMEXXX, type character, length 1
      local bRetVal
      bRetVal = true
   	try
      	if upper( right( this.destination, 4 ) ) # '.DBF'
         	this.destination += '.dbf'
         endif
         if empty( this.destination )
            throw new c4Err( 'Destination file not specified' )
         endif
         if empty( this.source )
         	throw new c4Err( 'Source rowset not specified' )
         endif
         this.fileObj.create( this.destination )
         this.fileObj.write( chr( 3 ) )	// version number
         // Date of last update, YYMMDD
         this.fileObj.write( chr( val( substr( str( year( date()),4 ),3,2))))
         this.fileObj.write( chr( month( date())))
         this.fileObj.write( chr( day( date())))
         // Number of records
         this.write32( 0 )
         // Number of bytes in header
         this.write16( 32 )
         // Number of bytes in record
         this.write16( 0 )
         // Fill byte 12-28 with zeros
         this.fileObj.write( replicate( chr( 0 ), 17 ) )
         // Language driver
         this.fileObj.write( chr( this.ldriver) )
         // Fill byte 30-31 with zeros
         this.fileObj.write( replicate( chr( 0 ), 2) )
         this.headerSize = 32
         this.recordSize = 1
         // Fieldname
         cTempName = 'XDELETEMEX'+chr(0)
         this.fileObj.write( cTempName )
         // Field type
         this.fileObj.write( 'C' )
         // 4 bytes reserved
         this.fileObj.write( replicate( chr( 0 ), 4) )
         // Field length
         this.fileObj.write( chr( 1 ) )
         // Decimals
         this.fileObj.write( chr( 0 ) )
         // 2 bytes reserved
         this.fileObj.write( replicate( chr( 0 ), 2) )
         // Work area ID
         this.fileObj.write( chr( 0 ) )
         // 10 bytes reserved
         this.fileObj.write( replicate( chr( 0 ), 10 ) )
         // Production .MDX field flag
         this.fileObj.write( chr( 0 ) )
         this.headerSize += 32
         this.recordSize ++
         // Field terminator
         this.fileObj.write( chr( 13 ) )
         this.headerSize ++
         this.fileObj.write( chr( 26 ) )
         this.fileObj.seek( 4 )
         // Number of records
         this.write32( this.recordNo )
         // Number of bytes in header
         this.write16( this.headerSize )
         // Number of bytes in record
         this.write16( this.recordSize )
         this.fileObj.close()
      catch ( exception e )
      	showError( e.message )
	      bRetVal = false
      finally
      	this.fileObj.close()
      endtry
      return bRetVal

   function addFields
   	local f, bRetVal
      bRetVal = true
      try
         f = 1
         this.aFields = new array()
         this.aNames = new array()
         for i = 1 to this.source.fields.size
            cTempName = this.db4Name( this.source.fields[i].fieldName )
            this.aFields.add( cTempName )
            do case
               case this.source.fields[ i ].fieldName == '_DBASELOCK'
                  this.aFields.size --
               case this.source.fields[i].type == 'CHARACTER'
                  this.aFields[this.aFields.size] += ' CHAR(' + ;
                     this.aFields[this.aFields.size].length+')'
                  this.aNames.add(this.source.fields[i].fieldName)
               case this.source.fields[i].type == 'NUMERIC'
                  this.aFields[this.aFields.size] += ' NUMERIC(' + ;
                     this.source.fields[i].length + ',' + ;
                     this.source.fields[i].decimalLength + ')'
                  this.aNames.add(this.source.fields[i].fieldName)
               case this.source.fields[i].type == 'DATE'
                  this.aFields[this.aFields.size] += ' DATE'
                  this.aNames.add(this.source.fields[i].fieldName)
               case this.source.fields[i].type == 'LOGICAL'
                  this.aFields[this.aFields.size] += ' BOOLEAN'
                  this.aNames.add(this.source.fields[i].fieldName)
               case this.source.fields[i].type == 'MEMO'
                  this.aFields[this.aFields.size] += ' BLOB(10,1)'
                  this.aNames.add(this.source.fields[i].fieldName)
               case this.source.fields[ i ].type == 'LONG'
                  this.aFields[this.aFields.size] += ' NUMERIC(20,0)'
                  this.aNames.add(this.source.fields[i].fieldName)
               case this.source.fields[i].type == 'DOUBLE'
                  this.aFields[this.aFields.size] += ' NUMERIC(' +;
                     this.source.fields[i].length + ',' + ;
                     this.source.fields[i].decimalLength + ')'
                  this.aNames.add(this.source.fields[i].fieldName)
               case this.source.fields[i].type == 'DATETIME'
                  this.aFields[this.aFields.size] += ' CHAR(19)'
                  this.aNames.add(this.source.fields[i].fieldName)
               case this.source.fields[i].type == 'AUTOINC'
                  this.aFields[this.aFields.size] += ' NUMERIC(8,0)'
                  this.aNames.add(this.source.fields[i].fieldName)
               case this.source.fields[i].type == 'FLOAT'
                  this.aFields[this.aFields.size] += ' NUMERIC(' +;
                     this.source.fields[i].length + ',' + ;
                     this.source.fields[i].decimalLength + ')'
                  this.aNames.add(this.source.fields[i].fieldName)
               otherwise
                  // ignore Ole and Binary fields
                  this.aFields.size --
            endcase
         endfor
         for i=1 to this.afields.size
            c = 'ALTER TABLE "'+this.destination+'" ' + 'ADD '+this.aFields[i]
            &c
         next
         c := 'ALTER TABLE "'+this.destination+'" ' + 'DROP XDELETEMEX'
         &C
      catch ( exception e )
      	showError( e.message )
	      bRetVal := false
      finally
      	this.fileObj.close()
      endtry
      return bRetVal


   function copyData
   	local bRetVal, f, nRecs, nCurrent
      bRetVal = true
   	try
         nCurrent = 0
         qSource = new query()
         qSource.sql = this.source.parent.sql
         qSource.database = this.source.parent.database
         qSource.active = true
         nRecs = qSource.rowset.count()
         qDest = new query()
         qDest.sql = 'select * from "'+this.destination+'"'
         qDest.active=true
         qSource.rowset.first()
         _app.cancelCopy = false
         this.progForm.pageno = 2
         do while not qSource.rowset.endOfSet and not _app.cancelCopy
            qDest.rowset.beginAppend()
            for i = 1 to this.aNames.size
               cName = this.aNames[i]
               xValue = qSource.rowset.fields[cName].value
               if qSource.rowset.fields[cName].type == 'DATETIME'
                  xValue = xValue.toLocaleString()
               endif
               qDest.rowset.fields[i].value = xValue
            endfor
            qDest.rowset.save()
            nCurrent ++
            this.progForm.progress1.value = 100 * nCurrent / nRecs
            qSource.rowset.next()
         enddo
      catch ( exception e )
      	showError( e )
	      bRetVal = false
      finally
      	this.fileObj.close()
         this.RecordsCopied := nCurrent
      endtry
      return bRetVal

   function db4name( cFieldName )
      // Convert VdB7 fieldname to dbf4 fieldname
      local cName4, f
      cName4 = upper( trim( ltrim( cFieldName ) ) )
      cName4 = left( cName4, 10 )
      do while at(' ', cName4)>0
         cName4= stuff( cName4, AT(' ', cName4), 1, '_')
      enddo
      f = 1
      do while this.aFields.scan(cName4) > 0
         cName4='FIELD'+transform( f, '@L 99999')
         cName4 := left( cName4 , ;
                        len( atrim( '' + f ) ) ) + ;
                        atrim( '' + f )
         f ++
      enddo
      return cName4

   function write16(num)
      // Write 16-bit number
      this.fileObj.write( chr( mod( num, 256 ) ) )
      this.fileObj.write( chr( int( num / 256 ) ) )
      return

   function Write32(num)
      // Write 32-bit number
      local byte0, byte1, byte2, byte3
      byte3 =int( num / ( 256 * 256 * 256 ) )
      num -= byte3 * 256 * 256 * 256
      byte2 = int( num / ( 256 * 256 ) )
      num -= byte2 * 256 * 256
      byte1 = int( num / ( 256 ))
      byte0 = num - (  byte1 * 256 )
      this.fileObj.write( chr( byte0 ) )
      this.fileObj.write( chr( byte1 ) )
      this.fileObj.write( chr( byte2 ) )
      this.fileObj.write( chr( byte3 ) )
      return

endclass

class progForm of FORM
   with (this)
      scaleFontBold = false
      height = 2.4545
      left = 49.5714
      top = 8.8182
      width = 41.2857
      text = "Copying data"
      topMost = true
      autoCenter = true
      mdi = false
      smallTitle = true
      sizeable = false
      moveable = false
      maximize = false
      minimize = false
   endwith

   this.TEXT1 = new TEXT(this)
   with (this.TEXT1)
      height = 1
      left = 0.4286
      top = 0.2273
      width = 39.8571
      text = "Copying data (Press Esc to cancel):"
      pageno = 2
   endwith

   this.PROGRESS1 = new PROGRESS(this)
   with (this.PROGRESS1)
      height = 1
      left = 0.4286
      top = 1.1364
      width = 39.8571
      pageno = 2
      value = 0
      rangeMin = 0
      rangeMax = 100
   endwith

   this.TEXT2 = new TEXT(this)
   with (this.TEXT2)
      height = 1
      left = 0.43
      top = 0.2273
      width = 35.1429
      text = "Creating file..."
   endwith

endclass

class c4Err( cMsg ) of exception
	this.message = cMsg
endclass
